/*
	BromScript - On Runtime Scripting Language.
    Copyright (C) 2012 - 2013  Alex Brouwer (Bromvlieg: http://brom.4o3.nl/)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "Converter.h"

using namespace Scratch;

namespace BromScript{
	class CVector{
public:
	double x;
	double y;
	double z;
};

	bool Converter::SetMember(Instance* bromscript, Variable* member, Variable* value, CString key){
		Userdata* ud = (Userdata*)member->Value;

		if (ud->Setter != null){
			ud->Setter(bromscript, ud->TypeData, value);
			return true;
		}

		if (ud->TypeID == MemberType::Bool && value->Type == Variable::VariableType::Bool){
			*(bool*)ud->TypeData = value->Value != null;
			return true;
		}else if (ud->TypeID == MemberType::Byte && value->Type == Variable::VariableType::Number){
			*(byte*)ud->TypeData = (byte)*(double*)value->Value;
			return true;
		}else if (ud->TypeID == MemberType::Int && value->Type == Variable::VariableType::Number){
			*(int*)ud->TypeData = (int)*(double*)value->Value;
			return true;
		}else if (ud->TypeID == MemberType::Short && value->Type == Variable::VariableType::Number){
			*(short*)ud->TypeData = (short)*(double*)value->Value;
			return true;
		}else if (ud->TypeID == MemberType::Float && value->Type == Variable::VariableType::Number){
			*(float*)ud->TypeData = (float)*(double*)value->Value;
			return true;
		}else if (ud->TypeID == MemberType::Double && value->Type == Variable::VariableType::Number){
			*(double*)ud->TypeData = *(double*)value->Value;
			return true;
		}else if (ud->TypeID == MemberType::Long && value->Type == Variable::VariableType::Number){
			*(long*)ud->TypeData = (long)*(double*)value->Value;
			return true;
		}else if (member->Type == value->Type && member->Type > MemberType::Userdata){
			memcpy(ud->TypeData, ((Userdata*)value->Value)->TypeData, ud->TypeSize);
			return true;
		}

		return false;
	}

	Variable* Converter::MemberToVariable(Instance* bromscript, Variable* member){
		Userdata* ud = (Userdata*)member->Value;
		
		if (ud->Getter != null){
			Variable* ret = ud->Getter(bromscript, ud->TypeData);
			if (ret == null) ret = bromscript->GC.RegisterVariable();
			else bromscript->GC.RegisterVariable(ret);

			return ret;
		}

		if (member->Type > MemberType::Bool)
			return member;

		Variable* var = bromscript->GC.RegisterVariable();
		var->IsCpp = member->IsCpp;
		var->Value = member->Value;
		var->Type = member->Type;
		if (member->CppData != null)
			var->CppData = member->CppData;

		switch (ud->TypeID)
		{
		case MemberType::Bool:
			var->Value = Converter::BoolToPointer(*(bool*)ud->TypeData);
			var->Type = Variable::VariableType::Bool;
			break;
		case MemberType::Double:
			var->Value = Converter::NumberToPointer(*(double*)ud->TypeData);
			var->Type = Variable::VariableType::Number;
			break;
		case MemberType::Byte:
			var->Value = Converter::NumberToPointer((double)*(byte*)ud->TypeData);
			var->Type = Variable::VariableType::Number;
			break;
		case MemberType::Int:
			var->Value = Converter::NumberToPointer((double)*(int*)ud->TypeData);
			var->Type = Variable::VariableType::Number;
			break;
		case MemberType::Short:
			var->Value = Converter::NumberToPointer((double)*(short*)ud->TypeData);
			var->Type = Variable::VariableType::Number;
			break;
		case MemberType::Float:
			var->Value = Converter::NumberToPointer((double)*(float*)ud->TypeData);
			var->Type = Variable::VariableType::Number;
			break;
		case MemberType::Long:
			var->Value = Converter::NumberToPointer((double)*(long*)ud->TypeData);
			var->Type = Variable::VariableType::Number;
			break;
		}

		return var;
	}

	Variable* Converter::ToVariable(Instance* bromscript, const char* key, BSFunction value){
		return Converter::ToVariable(bromscript, CString(key), value);
	}

	Variable* Converter::ToVariable(Instance* bromscript, CString key, BSFunction value){
		Function* varfunc = new Function(bromscript);
		varfunc->CppFunc = value;
		varfunc->Filename = "C++";
		varfunc->IsCpp = true;
		varfunc->Name = key;

		Variable* ret = new Variable();
		ret->Type = Variable::VariableType::Function;
		ret->Value = varfunc;
		ret->DeleteOnDestruct = false;

		return ret;
	}

	Variable* Converter::ToVariable(bool value){
		Variable* ret = new Variable();
		ret->Type = Variable::VariableType::Bool;
		ret->Value = Converter::BoolToPointer(value);
		return ret;
	}

	Variable* Converter::ToVariable(double value){
		Variable* ret = new Variable();
		ret->Type = Variable::VariableType::Number;
		ret->Value = Converter::NumberToPointer(value);
		return ret;
	}

	Variable* Converter::ToVariable(float value){
		Variable* ret = new Variable();
		ret->Type = Variable::VariableType::Number;
		ret->Value = Converter::NumberToPointer(value);
		return ret;
	}

	Variable* Converter::ToVariable(int value){
		Variable* ret = new Variable();
		ret->Type = Variable::VariableType::Number;
		ret->Value = Converter::NumberToPointer(value);
		return ret;
	}

	Variable* Converter::ToVariable(const char* value){
		Variable* ret = new Variable();
		ret->Type = Variable::VariableType::String;
		ret->Value = Converter::StringToPointer(value);
		return ret;
	}

	Variable* Converter::ToVariable(CString value){
		Variable* ret = new Variable();
		ret->Type = Variable::VariableType::String;
		ret->Value = Converter::StringToPointer(value);
		return ret;
	}

	Variable* Converter::ToVariable(Table* value){
		Variable* ret = new Variable();
		ret->Type = Variable::VariableType::Table;
		ret->Value = value;
		return ret;
	}

	void* Converter::NumberToPointer(double val){
		double& ret = *new double;
		ret = val;

		return &ret;
	}

	void* Converter::BoolToPointer(bool val){
		return (void*)(val ? 1 : null);
	}

	void* Converter::StringToPointer(CString val){
		return new CString(val);
	}

	CString Converter::TypeToString(Variable::VariableType type){
		switch(type){
		case Variable::VariableType::Bool: return "Bool";
		case Variable::VariableType::Class: return "Class";
		case Variable::VariableType::Function: return "Function";
		case Variable::VariableType::Null: return "Null";
		case Variable::VariableType::Number: return "Number";
		case Variable::VariableType::String: return "String";
		case Variable::VariableType::Table: return "Table";
		default:
			return "Userdata";
		}
	}

	CString Converter::TypeToString(Instance* bromscript, Variable::VariableType type){
		switch(type){
		case Variable::VariableType::Bool: return "Bool";
		case Variable::VariableType::Class: return "Class";
		case Variable::VariableType::Function: return "Function";
		case Variable::VariableType::Null: return "Null";
		case Variable::VariableType::Number: return "Number";
		case Variable::VariableType::String: return "String";
		case Variable::VariableType::Table: return "Table";
		default:
			for (int i = 0; i < bromscript->RegisteredUserdataTypes.Count; i++){
				if (bromscript->RegisteredUserdataTypes[i]->TypeID == type){
					return bromscript->RegisteredUserdataTypes[i]->Name;
				}
			}

			return "Userdata";
		}
	}

	CString Converter::TypeToString(Instance* bromscript, Variable* var){
		if (var == null) return "Null";

		switch(var->Type){
		case Variable::VariableType::Bool: return "Bool";
		case Variable::VariableType::Class: return "Class";
		case Variable::VariableType::Function: return "Function";
		case Variable::VariableType::Null: return "Null";
		case Variable::VariableType::Number: return "Number";
		case Variable::VariableType::String: return "String";
		case Variable::VariableType::Table:
			if (var->GetTable()->Get("__TYPE")->Type != Variable::VariableType::Null){
				return var->GetTable()->Get("__TYPE")->GetString();
			}

			return "Table";
		default:
			for (int i = 0; i < bromscript->RegisteredUserdataTypes.Count; i++){
				if (bromscript->RegisteredUserdataTypes[i]->TypeID == var->Type){
					return bromscript->RegisteredUserdataTypes[i]->Name;
				}
			}

			return "Userdata";
		}
	}

	CString Converter::ArithmaticToString(Misc::ArithmaticFuncs type){
		switch(type){
		case Misc::ArithmaticFuncs::Add: return "Add";
		case Misc::ArithmaticFuncs::And: return "And";
		case Misc::ArithmaticFuncs::Divide: return "Divide";
		case Misc::ArithmaticFuncs::Equal: return "Equal";
		case Misc::ArithmaticFuncs::GreaterOrEqual: return "GreaterOrEqual";
		case Misc::ArithmaticFuncs::GreaterThan: return "GreaterThan";
		case Misc::ArithmaticFuncs::LessOrEqual: return "LessOrEqual";
		case Misc::ArithmaticFuncs::LessThan: return "LessThan";
		case Misc::ArithmaticFuncs::Modulo: return "Modulo";
		case Misc::ArithmaticFuncs::Multiply: return "Multiply";
		case Misc::ArithmaticFuncs::NotEqual: return "NotEqual";
		case Misc::ArithmaticFuncs::Or: return "Or";
		case Misc::ArithmaticFuncs::Substract: return "Substract";
		case Misc::ArithmaticFuncs::BitwiseAnd: return "BitwiseAnd";
		case Misc::ArithmaticFuncs::BitwiseOr: return "BitwiseOr";
		case Misc::ArithmaticFuncs::BitwiseLeft: return "BitwiseLeft";
		case Misc::ArithmaticFuncs::BitwiseRight: return "BitwiseRight";
		default: return "Unknown";
		}
	}
}