/*
	BromScript - On Runtime Scripting Language.
    Copyright (C) 2012 - 2013  Alex Brouwer (Bromvlieg: http://brom.4o3.nl/)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef BROMSCRIPT_INSTANCE_INCLUDED
#define BROMSCRIPT_INSTANCE_INCLUDED

#include "SIF.h"

#include "Variable.h"
#include "CallStack.h"
#include "Function.h"
#include "ByteReader.h"
#include "GarbageCollector.h"
#include "Table.h"
#include "List.h"
#include "Userdata.h"
#include "Compiler.h"

namespace BromScript {
	class Instance
	{
	public:
		static int GetCount();

		Instance();
		~Instance();
		
		Scratch::CString CurrentIncludePath;
		Table* Globals;
		List<CallStack*> CurrentStack;
		List<Userdata*> RegisteredUserdataTypes;
		bool KillScriptThreaded;

		GarbageCollector GC;
		
		void LoadDefaultLibaries();
		Variable* DoString(const char* chunkname, const char* code);
		Variable* DoString(const char* chunkname, const char* code, int size);
		Variable* DoFile(const char* filename);
		Variable* DoCode(const char* filename, byte* code, int codelen);
		void DoDLL(const char* filename);

		void EnterStack(Function* func);
		void LeaveStack();

		void Error(Scratch::CString msg);

		void SetErrorCallback(ErrorFunction func);
		void RegisterFunction(Scratch::CString key, BSFunction func);
		void RegisterScope(Function* scope);

		Userdata* GetRegisteredUserdata(int typeID);
		Userdata* RegisterUserdata(Scratch::CString name, int typeID, int typesize, BSFunctionCtor ctor, BSFunctionDtor dtor);
		Variable* CreateUserdata(int typeID, void* ptr, bool calldtor);
		
		void RemoveVar(Scratch::CString key);
		void SetVar(Scratch::CString key, Variable* var);
		Variable* GetVar(Scratch::CString key);

		Function* GetCurrentFunction();
		CallStack* GetCurrentStack();
		int GetCurrentStackSize();

	private:
		List<Function*> ChunkScopes;
		List<BSModuleFunction> ModuleExits;

		bool SuppressErrors;

		int CurrentStackIndex;

		ErrorFunction ErrorCallback;
	};
}

#endif