/*
	BromScript - On Runtime Scripting Language.
    Copyright (C) 2012 - 2013  Alex Brouwer (Bromvlieg: http://brom.4o3.nl/)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SCRATCH_NO_GLOBFUNC
#define SCRATCH_NO_GLOBFUNC
#endif

#include "Scratch/Scratch.h"
#include "Variable.h"
#include "Converter.h"

using namespace Scratch;

namespace BromScript {
	extern int _vcount = 0;

	int Variable::GetCount(){
		return _vcount;
	}

	Variable::Variable(void)
	{
		_vcount++;
		this->IsCpp = false;
		this->ReadOnly = false;
		this->CppData = null;

		this->Value = null;
		this->Type = VariableType::Null;

		this->DeleteOnDestruct = true;

		this->Using = 0;
		this->RegisteredInGC = false;
	}

	Variable::~Variable(void)
	{
		_vcount--;

		if (!this->DeleteOnDestruct && this->Value == null)
			return;

		switch(this->Type){
		case VariableType::Null: return;
		case VariableType::Bool: return;
		case VariableType::Class: delete (BromScript::Function*)(this->Value); return;
		case VariableType::Function: delete (BromScript::Function*)(this->Value); return;
		case VariableType::String: delete (CString*)(this->Value); return;
		case VariableType::Number: delete (double*)(this->Value); return;
		case VariableType::Table: delete (BromScript::Table*)(this->Value); return;
		case VariableType::Userdata: return;
		default: // userdata
			delete (BromScript::Userdata*)(this->Value);
			return;
		}
	}

	const char* Variable::GetString(){
		if (this->Type != Variable::VariableType::String) return null;
		return *(CString*)this->Value;
	}
	
	double Variable::GetNumber(){
		if (this->Type != Variable::VariableType::Number) return 0;
		return *(double*)this->Value;
	}

	bool Variable::GetBool(){
		if (this->Type != Variable::VariableType::Bool) return false;
		return this->Value != null;
	}

	void* Variable::GetUserdata(){
		return ((BromScript::Userdata*)(this->Value))->TypeData;
	}
	
	Table* Variable::GetTable(){
		if (this->Type != Variable::VariableType::Table) return null;
		return (BromScript::Table*)this->Value;
	}

	Function* Variable::GetFunction(){
		if (this->Type != Variable::VariableType::Function) return null;
		return (BromScript::Function*)this->Value;
	}
}