﻿/*
	BromScript - On Runtime Scripting Language.
    Copyright (C) 2012 - 2013  Alex Brouwer (Bromvlieg: http://brom.4o3.nl/)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "BromScript.h"
#include <iostream>

//#define _CRTDBG_MAPALLOC
//#include <stdlib.h>
//#include <crtdbg.h>

using namespace BromScript;
using namespace Scratch;

BS_ERRORFUNCTION(errfunc){
	ArgumentData args;
	args.AddVariable(bsi->GetVar("Console")->GetTable()->Get("ColorRed"));
	bsi->GetVar("Console")->GetTable()->Get("SetForeColor")->GetFunction()->Run(&args, bsi->GetVar("Console"));

	printf("BromScript error!\n");

	for (int i = 0; i < stacksize; i++)
		printf("SID: %i, File: %s, Name: %s, Line: %d\n", i, stack[i].Filename, stack[i].Name, stack[i].LineNumber);

	printf("Message: %s\n", msg);
	bsi->GetVar("Console")->GetTable()->Get("SetDefaultColors")->GetFunction()->Run(&args, bsi->GetVar("Console"));
}

#define CENTUSERDATATYPE 9999
#define VECTORDATATYPE 9998

class CVector{
public:
	double x;
	double y;
	double z;
};

class CEnt{
public:
	CVector Pos;
	CVector Pos2;
	CString str;
};

BS_USERDATA_SET(CEntStrSet){
	*(CString*)ptr = CString(var->GetString());
}

BS_USERDATA_GET(CEntStrGet){
	CString str = (CString*)ptr;
	return Converter::ToVariable(str);
}

BS_FUNCTION_CTOR(CEntCtor){
	CEnt* test = new CEnt();
	test->str = "fawawf";
	return test;
}

BS_FUNCTION_DTOR(CEntDtor){
	delete (CEnt*)data;
}

BS_FUNCTION_CTOR(CVectorCtor){
	CVector* ct = new CVector();

	ct->x = args->GetNumber(0);
	ct->y = args->GetNumber(1);
	ct->z = args->GetNumber(2);

	return ct;
}

BS_FUNCTION_DTOR(CVectorDtor){
	delete (CVector*)data;
}

BS_FUNCTION(CVectorSetFunc){
	if (!args->CheckThisObject(VECTORDATATYPE, true)) return null;

	CVector* ct = (CVector*)args->GetThisObjectData();
	ct->x = args->GetNumber(0);
	ct->y = args->GetNumber(1);
	ct->z = args->GetNumber(2);

	return null;
}

BS_FUNCTION(CVectorGetXFunc){
	if (!args->CheckThisObject(VECTORDATATYPE, true)) return null;
	
	CVector* ct = (CVector*)args->GetThisObjectData();
	return Converter::ToVariable(ct->x);
}

BS_FUNCTION(CVectorGetYFunc){
	if (!args->CheckThisObject(VECTORDATATYPE, true)) return null;
	
	CVector* ct = (CVector*)args->GetThisObjectData();
	return Converter::ToVariable(ct->y);
}

BS_FUNCTION(CVectorGetZFunc){
	if (!args->CheckThisObject(VECTORDATATYPE, true)) return null;
	
	CVector* ct = (CVector*)args->GetThisObjectData();
	return Converter::ToVariable(ct->z);
}

BS_FUNCTION(CVectorAdd){
	if (!args->CheckType(0, VECTORDATATYPE, true)) return null;
	if (!args->CheckType(1, VECTORDATATYPE, true)) return null;

	CVector* ct1 = (CVector*)args->GetUserdata(0, VECTORDATATYPE);
	CVector* ct2 = (CVector*)args->GetUserdata(1, VECTORDATATYPE);

	CVector* robj = new CVector();
	robj->x = ct1->x + ct2->x;
	robj->y = ct1->y + ct2->y;
	robj->z = ct1->z + ct2->z;

	Variable* ret = new Variable();
	ret->Type = (Variable::VariableType)VECTORDATATYPE;

	Userdata* classdata = bsi->GetRegisteredUserdata(VECTORDATATYPE)->Copy();
	classdata->TypeData = robj;
	classdata->CallDTor = true;
	ret->Value = classdata;

	return ret;
}

BS_FUNCTION(CVectorMul){
	if (!args->CheckType(0, VECTORDATATYPE, true)) return null;
	if (!args->CheckType(1, Variable::VariableType::Number, true)) return null;
	
	CVector* ct1 = (CVector*)args->GetUserdata(0, VECTORDATATYPE);

	int mulnum = args->GetNumber(1);

	CVector* robj = new CVector();
	robj->x = ct1->x * mulnum;
	robj->y = ct1->y * mulnum;
	robj->z = ct1->z * mulnum;

	Variable* ret = new Variable();
	ret->Type = (Variable::VariableType)VECTORDATATYPE;

	Userdata* classdata = bsi->GetRegisteredUserdata(VECTORDATATYPE)->Copy();
	classdata->TypeData = robj;
	classdata->CallDTor = true;

	ret->Value = classdata;

	return ret;
}

int main(){
	system("Title C++ BromScript");

	Instance* bsi = new Instance();
	bsi->LoadDefaultLibaries();
	
	Userdata* vd = bsi->RegisterUserdata("CVector", VECTORDATATYPE, sizeof(CVector), CVectorCtor, CVectorDtor);
	vd->RegisterMember("x", offsetof(CVector, x), MemberType::Double);
	vd->RegisterMember("y", offsetof(CVector, y), MemberType::Double);
	vd->RegisterMember("z", offsetof(CVector, z), MemberType::Double);
	vd->RegisterFunction("Set", CVectorSetFunc);
	vd->RegisterFunction("GetX", CVectorGetXFunc);
	vd->RegisterFunction("GetY", CVectorGetYFunc);
	vd->RegisterFunction("GetZ", CVectorGetZFunc);
	
	vd->RegisterOperator(Misc::ArithmaticFuncs::Add, CVectorAdd);
	vd->RegisterOperator(Misc::ArithmaticFuncs::Multiply, CVectorMul);

	Userdata* ud = bsi->RegisterUserdata("CEnt", CENTUSERDATATYPE, sizeof(CEnt), CEntCtor, CEntDtor);
	ud->RegisterMember("Pos", offsetof(CEnt, Pos), VECTORDATATYPE);
	ud->RegisterMember("Pos2", offsetof(CEnt, Pos2), VECTORDATATYPE);
	ud->RegisterMember("str", offsetof(CEnt, str), MemberType::String, CEntStrSet, CEntStrGet);

	bsi->SetErrorCallback(errfunc);

	{ // leak scope :V
		List<CString> tests;
		/*tests.Add("function a(){ print(\"a\") return true }");
		tests.Add("function b(){ print(\"b\") return true }");
		tests.Add("print(a() && b())");
		tests.Add("GC.Run()");
		tests.Add("print(2/2*1+1*5*-1 == -4)");
		tests.Add("print(5 - 5 % 2)");
		tests.Add("print(\"Lol, what?\" / 0.1)");
		tests.Add("tbl = {} function tbl.test(a){ print(this) }");
		tests.Add("tbl.test()");*/
		/*tests.Add("function a(){\r\nreturn a\r\n} print(a())");
		tests.Add("ent = new CEnt()");
		tests.Add("print(ent.str)");
		tests.Add("ent.str = \"nigger\"");
		tests.Add("print(ent.str)");
		tests.Add("foreach (k, v, ent) print(k, v)");*/

		for (int i = 0; i < tests.Count; i++)
			bsi->DoString("test", tests[i]);
	}

	// you can find usercase file at http://4o3.nl/c8a0b.txt
	bsi->DoFile("D:\\AutoSync\\Programming\\C#\\Brom_Compiler\\BromScriptRunner\\bin\\Debug\\test.bs");

	delete bsi;
	printf("\n\nAaaaaand we're done here.\n");

	//_CrtDumpMemoryLeaks();
	std::cin.get();
	return 0;
}